"""Handling of .SRCINFO files."""

from __future__ import annotations

from io import StringIO
from logging import debug
from pathlib import Path

from orjson import OPT_APPEND_NEWLINE, OPT_INDENT_2, OPT_SORT_KEYS, dumps
from pydantic import BaseModel, HttpUrl, ValidationError, conlist, constr

from repod.common.enums import FieldTypeEnum
from repod.common.models import (
    Arch,
    Backup,
    CheckDepends,
    Conflicts,
    Depends,
    Epoch,
    Groups,
    License,
    MakeDepends,
    OptDepends,
    Options,
    PkgBase,
    PkgDesc,
    PkgName,
    PkgRel,
    PkgVer,
    Provides,
    Replaces,
    Url,
)
from repod.common.regex import (
    ARCHITECTURE,
    B2,
    CK,
    MD5,
    PGP_KEY_ID,
    SHA1,
    SHA224,
    SHA256,
    SHA384,
    SHA512,
)
from repod.errors import FileParserError
from repod.files.common import read_text_from_file

SRCINFO_ASSIGNMENTS: dict[str, tuple[str, FieldTypeEnum]] = {
    "arch": ("arch", FieldTypeEnum.STRING),
    "backup": ("backup", FieldTypeEnum.STRING_LIST),
    "b2sums": ("b2sums", FieldTypeEnum.STRING_LIST),
    "changelog": ("changelog", FieldTypeEnum.STRING),
    "checkdepends": ("checkdepends", FieldTypeEnum.STRING_LIST),
    "cksums": ("cksums", FieldTypeEnum.STRING_LIST),
    "conflicts": ("conflicts", FieldTypeEnum.STRING_LIST),
    "depends": ("depends", FieldTypeEnum.STRING_LIST),
    "epoch": ("epoch", FieldTypeEnum.INT),
    "groups": ("groups", FieldTypeEnum.STRING_LIST),
    "install": ("install", FieldTypeEnum.STRING),
    "license": ("license", FieldTypeEnum.STRING_LIST),
    "makedepends": ("makedepends", FieldTypeEnum.STRING_LIST),
    "md5sum": ("md5sums", FieldTypeEnum.STRING_LIST),
    "noextract": ("noextract", FieldTypeEnum.STRING_LIST),
    "optdepends": ("optdepends", FieldTypeEnum.STRING_LIST),
    "options": ("options", FieldTypeEnum.STRING_LIST),
    "pkgbase": ("pkgbase", FieldTypeEnum.STRING),
    "pkgdesc": ("pkgdesc", FieldTypeEnum.STRING),
    "pkgname": ("pkgname", FieldTypeEnum.STRING),
    "pkgrel": ("pkgrel", FieldTypeEnum.STRING),
    "pkgver": ("pkgver", FieldTypeEnum.STRING),
    "provides": ("provides", FieldTypeEnum.STRING_LIST),
    "replaces": ("replaces", FieldTypeEnum.STRING_LIST),
    "sha1sums": ("sha1sums", FieldTypeEnum.STRING_LIST),
    "sha224sums": ("sha224sums", FieldTypeEnum.STRING_LIST),
    "sha256sums": ("sha256sums", FieldTypeEnum.STRING_LIST),
    "sha384sums": ("sha384sums", FieldTypeEnum.STRING_LIST),
    "sha512sums": ("sha512sums", FieldTypeEnum.STRING_LIST),
    "source": ("source", FieldTypeEnum.STRING_LIST),
    "url": ("url", FieldTypeEnum.STRING),
    "validpgpkeys": ("validpgpkeys", FieldTypeEnum.STRING_LIST),
}


def parse_pairs(line: str, separator: str = " = ") -> tuple[str, str, FieldTypeEnum]:
    """Parse key-value pairs from a line of text.

    The line of text represents the data contained in a .SRCINFO file.
    Keys are resolved based on SRCINFO_ASSIGMENTS.

    Parameters
    ----------
    line: str
        A line of text to be parsed
    separator: str
        A separator string by which to split the line of text

    Raises
    ------
    FileParserError
        If an invalid line of text is provided.

    Returns
    -------
    tuple[str, str, FieldTypeEnum]
        A tuple of two strings and a member of FieldTypeEnum, which represent key, value and field type extracted from
        the line of text
    """
    debug(f"Parsing: {line}")
    line = line.strip()

    try:
        extracted_key, value = [x.strip() for x in line.strip().split(separator, 1)]

        assignment_key = SRCINFO_ASSIGNMENTS.get(extracted_key)
        if assignment_key is None:
            raise FileParserError(
                f"An error occured parsing the .SRCINFO line '{line}'! "
                f"The key {extracted_key} can not be found in the type assignments for .SRCINFO keywords."
            )
        key = assignment_key[0]
        field_type = assignment_key[1]
    except ValueError as e:
        raise FileParserError(f"An error occurred while trying to parse the .SRCINFO line {line}\n{e}")

    return key, value, field_type


def pairs_to_entries(
    key: str,
    value: str,
    field_type: FieldTypeEnum,
    entries: dict[str, int | str | list[str]],
) -> None:
    """Append key value-pairs to a dict.

    Values are cast based on their provided field types.
    Nested values for xdata are understood and are initialized as their target types (e.g. PkgType).

    Parameters
    ----------
    key: str
        The resolved key
    value: str
        The extracted value
    field_type: FieldTypeEnum
        A member of FieldTypeEnum based upon which value is cast to a specific type
    entries: dict[str, Any]
        The dict to which the key-value pairs are added

    Raises
    ------
    RuntimeError
        If an invalid key/ field type combination is encountered
    """
    debug(f"Attempting to add key {key} and value {value} of field type {field_type.value} to a dict...")
    match field_type:
        case FieldTypeEnum.INT if entries.get(key):
            raise FileParserError(
                "Failed parsing .SRCINFO data: "
                f"The key-value pair '{key}' => '{entries.get(key)}' exists already, unable to use value '{value}'!"
            )
        case FieldTypeEnum.INT:
            try:
                entries[key] = int(value)
            except ValueError:
                raise FileParserError(
                    "Failed parsing .SRCINFO data: " f"The value '{value}' of key {key} can not be converted to int!"
                )
        case FieldTypeEnum.STRING if entries.get(key):
            raise FileParserError(
                "Failed parsing .SRCINFO data: "
                f"The key-value pair '{key}' => '{entries.get(key)}' exists already, unable to use value '{value}'!"
            )
        case FieldTypeEnum.STRING:
            entries[key] = str(value)
        case FieldTypeEnum.STRING_LIST:
            entry = entries.get(key)
            if entry is not None and isinstance(entry, list):
                entry.append(str(value))
            else:
                entries[key] = [str(value)]
        case _:
            raise FileParserError(
                "An invalid field type has been encountered while attempting to read a .SRCINFO file."
            )


class B2Sums(BaseModel):
    """The representation of an optional list of blake2 checksum or 'SKIP' strings in a .SRCINFO file.

    Attributes
    ----------
    b2sums: list[str] | None
        An optional list of blake2 checksum or 'SKIP' strings
    """

    b2sums: list[constr(pattern=f"^({B2}|SKIP)$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class CkSums(BaseModel):
    """The representation of an optional list of CRC-32 checksum or 'SKIP' strings in a .SRCINFO file.

    Attributes
    ----------
    cksums: list[str] | None
        An optional list of CRC-32 checksum or 'SKIP' strings
    """

    cksums: list[constr(pattern=f"^({CK}|SKIP)$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class Changelog(BaseModel):
    """The representation of an optional changelog file string in a .SRCINFO file.

    Attributes
    ----------
    changelog: str | None
        An optional changelog file string
    """

    changelog: str | None = None


class Install(BaseModel):
    """The representation of an optional install file string in a .SRCINFO file.

    Attributes
    ----------
    install: str | None
        An optional .install file string
    """

    install: str | None = None


class Md5Sums(BaseModel):
    """The representation of an optional list of MD5 checksum or 'SKIP' strings in a .SRCINFO file.

    Attributes
    ----------
    md5sums: list[str] | None
        An optional list of MD5 checksum or 'SKIP' strings
    """

    md5sums: list[constr(pattern=f"^({MD5}|SKIP)$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class Noextract(BaseModel):
    """The representation of an optional list of file strings in a .SRCINFO file marked for no extraction.

    Attributes
    ----------
    noextract: list[str] | None
        An optional list of file string marked for no extraction
    """

    noextract: list[str] | None = None


class Sha1Sums(BaseModel):
    """The representation of an optional list of SHA-1 checksum or 'SKIP' strings in a .SRCINFO file.

    Attributes
    ----------
    sha1sums: list[str] | None
        An optional list of SHA-1 checksum or 'SKIP' strings
    """

    sha1sums: list[constr(pattern=f"^({SHA1}|SKIP)$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class Sha224Sums(BaseModel):
    """The representation of an optional list of SHA-224 checksum or 'SKIP' strings in a .SRCINFO file.

    Attributes
    ----------
    sha224sums: list[str] | None
        An optional list of SHA-224 checksum or 'SKIP' strings
    """

    sha224sums: list[constr(pattern=f"^({SHA224}|SKIP)$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class Sha256Sums(BaseModel):
    """The representation of an optional list of SHA-256 checksum or 'SKIP' strings in a .SRCINFO file.

    Attributes
    ----------
    sha256sums: list[str] | None
        An optional list of SHA-256 checksum or 'SKIP' strings
    """

    sha256sums: list[constr(pattern=f"^({SHA256}|SKIP)$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class Sha384Sums(BaseModel):
    """The representation of an optional list of SHA-384 checksum or 'SKIP' strings in a .SRCINFO file.

    Attributes
    ----------
    sha384sums: list[str] | None
        An optional list of SHA-384 checksum or 'SKIP' strings
    """

    sha384sums: list[constr(pattern=f"^({SHA384}|SKIP)$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class Sha512Sums(BaseModel):
    """The representation of an optional list of SHA-512 checksum or 'SKIP' strings in a .SRCINFO file.

    Attributes
    ----------
    sha512sums: list[str] | None
        An optional list of SHA-512 checksum or 'SKIP' strings
    """

    sha512sums: list[constr(pattern=f"^({SHA512}|SKIP)$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class Source(BaseModel):
    """The representation of an optional list of source (file or remote artifact) strings in a .SRCINFO file.

    Attributes
    ----------
    source: list[str] | None
        An optional list of source (file or remote artifact) strings
    """

    source: list[str] | None = None


class ValidPGPKeys(BaseModel):
    """The representation of an optional list of valid PGP key ID strings in a .SRCINFO file.

    Attributes
    ----------
    validpgpkeys: list[str] | None
        An optional list of PGP key ID strings
    """

    validpgpkeys: list[constr(pattern=PGP_KEY_ID)] | None = None  # type: ignore[valid-type]  # noqa: F722


class OptionalArch(BaseModel):
    """The representation of an optional architecture string in a .SRCINFO file.

    Attributes
    ----------
    arch: str | None
        An optional architecture string of a package
    """

    arch: constr(pattern=f"^{ARCHITECTURE}$") | None = None  # type: ignore[valid-type]  # noqa: F722


class OptionalLicense(BaseModel):
    """A model describing a single 'license' attribute.

    Attributes
    ----------
    license: list[str] | None
        An optional list of license identifier strings, that describe the license(s) of a package
    """

    license: list[str]


class OptionalPkgDesc(BaseModel):
    """The representation of an optional package description string in a .SRCINFO file.

    Attributes
    ----------
    pkgdesc: str | None
        An optional package description string
    """

    pkgdesc: str | None = None


class OptionalUrl(BaseModel):
    """The representation of an optional HTTP URL in a .SRCINFO file.

    Attributes
    ----------
    url: HttpUrl | None
        An optional URL string representing the upstream project of a package
    """

    url: HttpUrl | None = None


class PkgBaseSection(BaseModel):
    """The representation of a pkgbase section in a .SRCINFO file.

    Refer to specific implementations (e.g. PkgBaseSectionV1) for attributes.
    """

    pass


class PkgBaseSectionV1(
    Arch,
    B2Sums,
    Backup,
    Changelog,
    CheckDepends,
    CkSums,
    Conflicts,
    Depends,
    Epoch,
    Groups,
    Install,
    License,
    MakeDepends,
    Md5Sums,
    Noextract,
    OptDepends,
    Options,
    PkgBase,
    PkgBaseSection,
    PkgDesc,
    PkgRel,
    PkgVer,
    Provides,
    Replaces,
    Sha1Sums,
    Sha224Sums,
    Sha256Sums,
    Sha384Sums,
    Sha512Sums,
    Source,
    Url,
    ValidPGPKeys,
):
    """The representation of a pkgbase section in a .SRCINFO file (version 1).

    Attributes
    ----------
    arch: str
        A package architecture
    b2sums: list[str] | None
        An optional list of blake2 checksum or 'SKIP' strings
    backup: list[str] | None
        An optional list of relative file names to backup
    changelog: str | None
        An optional changelog file string
    checkdepends: list[str] | None
        An optional list of check dependencies
    cksums: list[str] | None
        An optional list of CRC-32 checksum or 'SKIP' strings
    conflicts: list[str] | None
        An optional list of packages conflicting with a package
    depends: list[str] | None
        An optional list of depdendency package names
    epoch: PositiveInt | None
        An optional positive integer representing the epoch of a package
    groups: list[str] | None
        An optional list of groups that a package belongs to
    install: str | None
        An optional install file string
    license: list[str]
        A list of license identifier strings, that describe the license(s) of a package
    makedepends: list[str] | None
        An optional list of package names required for building a package
    md5sums: list[str] | None
        An optional list of MD5 checksum or 'SKIP' strings
    noextract: list[str] | None
        An optional list of file string marked for no extraction
    optdepends: list[str] | None
        And optional list of optional dependencies of a package
    options: list[str] | None
        An optional list of strings representing makepkg.conf OPTIONS used during the creation of a package
    pkgbase: str
        A string representing the pkgbase of a package
    pkgdesc: str
        A string used as package description
    pkgrel: str
        A string representing the pkgrel (package release version) of a package
    pkgver: str
        A string representing the pkgver (upstream package version) of a package
    provides: list[str] | None
        An optional list of packages, that are virtually provided by a package
    replaces: list[str] | None
        An optional list of packages, that are virtually replaced by a package
    sha1sums: list[str] | None
        An optional list of SHA-1 checksum or 'SKIP' strings
    sha224sums: list[str] | None
        An optional list of SHA-224 checksum or 'SKIP' strings
    sha256sums: list[str] | None
        An optional list of SHA-256 checksum or 'SKIP' strings
    sha384sums: list[str] | None
        An optional list of SHA-384 checksum or 'SKIP' strings
    sha512sums: list[str] | None
        An optional list of SHA-512 checksum or 'SKIP' strings
    source: list[str] | None
        An optional list of source (file or remote artifact) strings
    url: str
        A URL string representing the upstream project of a package
    validpgpkeys: list[str] | None
        An optional list of PGP key ID strings
    """

    pass


class PkgNameSection(BaseModel):
    """The representation of a pkgname section in a .SRCINFO file (version 1).

    Refer to specific implementations (e.g. PkgNameSectionV1) for attributes.
    """

    pass


class PkgNameSectionV1(
    Backup,
    Changelog,
    Conflicts,
    Depends,
    Groups,
    Install,
    OptDepends,
    OptionalArch,
    OptionalPkgDesc,
    OptionalUrl,
    PkgName,
    PkgNameSection,
    Provides,
    Replaces,
):
    """The representation of a pkgname section in a .SRCINFO file (version 1).

    Attributes
    ----------
    arch: str | None
        An optional architecture string of a package
    backup: list[str] | None
        An optional list of relative file names to backup
    changelog: str | None
        An optional changelog file string
    conflicts: list[str] | None
        An optional list of packages conflicting with a package
    depends: list[str] | None
        An optional list of depdendency package names
    groups: list[str] | None
        An optional list of groups that a package belongs to
    install: str | None
        An optional .install file string
    optdepends: list[str] | None
        And optional list of optional dependencies of a package
    pkgdesc: str | None
        An optional package description string
    pkgname: str
        A string representing the pkgname of a package
    provides: list[str] | None
        An optional list of packages, that are virtually provided by a package
    replaces: list[str] | None
        An optional list of packages, that are virtually replaced by a package
    url: HttpUrl | None
        An optional URL string representing the upstream project of a package
    """

    pass


class SrcInfo(BaseModel):
    """The representation of a .SRCINFO file.

    Refer to specific implementations (e.g. SrcInfoV1) for attributes.
    """

    @classmethod
    def from_file(cls, data: Path | StringIO) -> SrcInfo:
        """Create a SrcInfo from a file path or StringIO.

        Parameters
        ----------
        data: Path | StringIO
            A Path or StringIO representing .SRCINFO file data

        Returns
        -------
        SrcInfo
            An instance of SrcInfo
        """
        if isinstance(data, (str, Path)):
            data = read_text_from_file(path=data)

        pkgbase: dict[str, int | str | list[str]] = {}
        pkgnames: list[dict[str, int | str | list[str]]] = []
        current_dict: dict[str, int | str | list[str]] = {}
        current_pkgname = 0
        for line in [line for line in data if line.strip() and not line.startswith("#")]:
            key, value, field_type = parse_pairs(line=line)

            match key:
                case "pkgbase":
                    if pkgbase:
                        raise FileParserError("Can not have more than one pkgbase per .SRCINFO file!")

                    current_dict = pkgbase
                case "pkgname":
                    if not pkgbase:
                        raise FileParserError("The pkgname section can not be the first section in a .SRCINFO file!")

                    pkgnames += [{}]
                    current_pkgname = len(pkgnames) - 1
                    current_dict = pkgnames[current_pkgname]
                case _:
                    if not pkgbase:
                        raise FileParserError(
                            f"The key-value pair '{key}' => '{value}' can not be the first entry in a .SRCINFO file!"
                        )

            pairs_to_entries(key=key, value=value, field_type=field_type, entries=current_dict)

        try:
            return SrcInfoV1(
                pkgbase=PkgBaseSectionV1.model_validate(pkgbase),  # pyright: ignore
                pkgnames=[PkgNameSectionV1.model_validate(pkgname) for pkgname in pkgnames],  # pyright: ignore
            )
        except ValidationError as e:
            raise FileParserError(e)


class SrcInfoV1(SrcInfo):
    """The representation of a .SRCINFO file (version 1).

    Attributes
    ----------
    pkgbase: PkgBaseSectionV1
        A PkgBaseSectionV1 describing the pkgbase section of the .SRCINFO file
    pkgnames: list[PkgNameSectionV1]
        A list of PkgNameSectionV1 (with at least one item) describing the pkgname sections of the .SRCINFO file
    """

    pkgbase: PkgBaseSectionV1
    pkgnames: conlist(PkgNameSectionV1, min_length=1)  # type: ignore[valid-type]


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema of selected pydantic models to an output directory.

    Parameters
    ----------
    output: Path
        A path to which to output the JSON schema files

    Raises
    ------
    RuntimeError
        If output is not an existing directory
    """
    classes = [SrcInfoV1, PkgBaseSectionV1, PkgNameSectionV1]

    if isinstance(output, str):
        output = Path(output)

    if not output.exists():
        raise RuntimeError(f"The output directory {output} must exist!")

    for class_ in classes:
        with open(output / f"{class_.__name__}.json", "wb") as f:
            f.write(
                dumps(
                    class_.model_json_schema(),  # type: ignore[attr-defined]
                    option=OPT_INDENT_2 | OPT_APPEND_NEWLINE | OPT_SORT_KEYS,
                )
            )



"""Package repository handling for repod."""

from repod.repo.package.repofile import RepoFile  # noqa: F401
from repod.repo.package.syncdb import Files  # noqa: F401
from repod.repo.package.syncdb import PackageDesc  # noqa: F401
from repod.repo.package.syncdb import RepoDbMemberTypeEnum  # noqa: F401
from repod.repo.package.syncdb import RepoDbTypeEnum  # noqa: F401
from repod.repo.package.syncdb import SyncDatabase  # noqa: F401
from repod.repo.package.syncdb import export_schemas  # noqa: F401
from repod.repo.package.syncdb import get_desc_json_field_type  # noqa: F401
from repod.repo.package.syncdb import get_desc_json_keys  # noqa: F401
from repod.repo.package.syncdb import get_desc_json_name  # noqa: F401
from repod.repo.package.syncdb import get_files_json_field_type  # noqa: F401
from repod.repo.package.syncdb import get_files_json_keys  # noqa: F401
from repod.repo.package.syncdb import get_files_json_name  # noqa: F401



"""Implementation of PGP based verification."""

from __future__ import annotations

from abc import ABC, abstractmethod
from logging import info
from pathlib import Path

from repod.commands import run_command


class PGPVerifier(ABC):
    """An abstract base class implementing PGP verification."""

    @abstractmethod
    def verify(self: PGPVerifier, package: Path, signature: Path) -> bool:
        """Verify a package and its accompanying signature.

        Parameters
        ----------
        package: Path
            The path to a package file
        signature: Path
            The path to a PGP signature for package

        Returns
        -------
        bool
            True if the signature can be verified, False otherwise
        """
        pass  # pragma: nocover


class PacmanKeyVerifier(PGPVerifier):
    """A class implementing PGP verification using pacman-key."""

    def verify(self: PGPVerifier, package: Path, signature: Path) -> bool:
        """Verify the detached PGP signature of a package file using pacman-key --verify.

        Parameters
        ----------
        package: Path
            The path to a package file
        signature: Path
            The path to a PGP signature for package

        Returns
        -------
        bool
            True if the signature can be verified, False otherwise
        """
        result = run_command(
            cmd=["pacman-key", "--verify", f"{signature}"],
            quiet=True,
        )

        if result.returncode != 0:
            info(f"The package file {package} could not be verified using the signature {signature}!\n{result.stderr}")
            return False

        return True



"""Module with configuration defaults for settings.

The defaults for locations and directories are provided as dicts which use SettingsTypeEnum as keys to distinguish
between different ways of running repod.
"""

from pathlib import Path

from orjson import OPT_APPEND_NEWLINE, OPT_INDENT_2, OPT_SORT_KEYS
from xdg.BaseDirectory import xdg_config_home, xdg_state_home

from repod.common.enums import ArchitectureEnum, CompressionTypeEnum, SettingsTypeEnum

DEFAULT_ARCHITECTURE = ArchitectureEnum.ANY
DEFAULT_BUILD_REQUIREMENTS_EXIST: bool = True
DEFAULT_DATABASE_COMPRESSION = CompressionTypeEnum.GZIP
DEFAULT_NAME = "default"

ORJSON_OPTION = OPT_INDENT_2 | OPT_APPEND_NEWLINE | OPT_SORT_KEYS

SETTINGS_LOCATION = {
    SettingsTypeEnum.SYSTEM: Path("/etc/repod.conf"),
    SettingsTypeEnum.USER: Path(xdg_config_home + "/repod/repod.conf"),
}
SETTINGS_OVERRIDE_LOCATION = {
    SettingsTypeEnum.SYSTEM: Path("/etc/repod.conf.d/"),
    SettingsTypeEnum.USER: Path(xdg_config_home + "/repod/repod.conf.d/"),
}

MANAGEMENT_REPO_BASE = {
    SettingsTypeEnum.SYSTEM: Path("/var/lib/repod/management/"),
    SettingsTypeEnum.USER: Path(xdg_state_home + "/repod/management/"),
}
PACKAGE_POOL_BASE = {
    SettingsTypeEnum.SYSTEM: Path("/var/lib/repod/data/pool/package/"),
    SettingsTypeEnum.USER: Path(xdg_state_home + "/repod/data/pool/package/"),
}
PACKAGE_REPO_BASE = {
    SettingsTypeEnum.SYSTEM: Path("/var/lib/repod/data/repo/package/"),
    SettingsTypeEnum.USER: Path(xdg_state_home + "/repod/data/repo/package/"),
}
SOURCE_POOL_BASE = {
    SettingsTypeEnum.SYSTEM: Path("/var/lib/repod/data/pool/source/"),
    SettingsTypeEnum.USER: Path(xdg_state_home + "/repod/data/pool/source/"),
}
SOURCE_REPO_BASE = {
    SettingsTypeEnum.SYSTEM: Path("/var/lib/repod/data/repo/source/"),
    SettingsTypeEnum.USER: Path(xdg_state_home + "/repod/data/repo/source/"),
}
PACKAGE_ARCHIVE_DIR = {
    SettingsTypeEnum.SYSTEM: Path("/var/lib/repod/archive/package/"),
    SettingsTypeEnum.USER: Path(xdg_state_home + "/repod/archive/package/"),
}
SOURCE_ARCHIVE_DIR = {
    SettingsTypeEnum.SYSTEM: Path("/var/lib/repod/archive/source/"),
    SettingsTypeEnum.USER: Path(xdg_state_home + "/repod/archive/source/"),
}



"""Handling of .BUILDINFO files."""

from __future__ import annotations

from io import StringIO
from pathlib import Path
from re import fullmatch
from typing import Any

from orjson import OPT_APPEND_NEWLINE, OPT_INDENT_2, OPT_SORT_KEYS, dumps
from pydantic import BaseModel, NonNegativeInt, constr, field_validator, model_validator

from repod.common.enums import ArchitectureEnum, FieldTypeEnum
from repod.common.models import (
    Options,
    Packager,
    PkgBase,
    PkgName,
    SchemaVersionV1,
    SchemaVersionV2,
)
from repod.common.regex import (
    ARCHITECTURE,
    BUILDENVS,
    EPOCH,
    PACKAGE_NAME,
    PKGREL,
    SHA256,
    VERSION,
)
from repod.errors import RepoManagementError

BUILDINFO_ASSIGNMENTS: dict[str, tuple[str, FieldTypeEnum]] = {
    "builddate": ("builddate", FieldTypeEnum.INT),
    "builddir": ("builddir", FieldTypeEnum.STRING),
    "buildenv": ("buildenv", FieldTypeEnum.STRING_LIST),
    "buildtool": ("buildtool", FieldTypeEnum.STRING),
    "buildtoolver": ("buildtoolver", FieldTypeEnum.STRING),
    "format": ("schema_version", FieldTypeEnum.INT),
    "installed": ("installed", FieldTypeEnum.STRING_LIST),
    "options": ("options", FieldTypeEnum.STRING_LIST),
    "packager": ("packager", FieldTypeEnum.STRING),
    "pkgarch": ("pkgarch", FieldTypeEnum.STRING),
    "pkgbase": ("pkgbase", FieldTypeEnum.STRING),
    "pkgbuild_sha256sum": ("pkgbuild_sha256sum", FieldTypeEnum.STRING),
    "pkgname": ("pkgname", FieldTypeEnum.STRING),
    "pkgver": ("pkgver", FieldTypeEnum.STRING),
    "startdir": ("startdir", FieldTypeEnum.STRING),
}


class BuildDate(BaseModel):
    """A build date in seconds since the epoch.

    Attributes
    ----------
    builddate: NonNegativeInt
        A number >= 0
    """

    builddate: NonNegativeInt


class BuildDir(BaseModel):
    """An absolute build directory.

    Attributes
    ----------
    builddir: str
        A string representing an absolute directory
    """

    builddir: str

    @field_validator("builddir")
    @classmethod
    def validate_builddir(cls, builddir: str) -> str:
        """Validate the builddir attribute.

        The builddir attribute may not contain strings that represent absolute Paths or Paths in the home directory

        Parameters
        ----------
        builddir: str
            A string, representing a path

        Returns
        -------
        str
            A validated string, representing an absolute Path
        """
        path = Path(builddir)
        if not path.is_absolute():
            raise ValueError(f"A relative path as builddir is not valid: {path}")

        return builddir


class BuildEnv(BaseModel):
    """A list of build environment options.

    For valid values refer to BUILDENV in https://man.archlinux.org/man/makepkg.conf.5#OPTIONS

    Attributes
    ----------
    buildenv: list[str]
        A list of strings as described by makepkg.conf's BUILDENV option
    """

    buildenv: list[constr(pattern=rf"^{BUILDENVS}$")]  # type: ignore[valid-type]  # noqa: F722


class BuildTool(BaseModel):
    """The build tool used to create a package.

    Attributes
    ----------
    buildtool: str
        The package name of the build tool used to create a package
    """

    buildtool: constr(pattern=rf"^{PACKAGE_NAME}$")  # type: ignore[valid-type]  # noqa: F722


class BuildToolVer(BaseModel):
    """The package version of the build tool used to create a package.

    Attributes
    ----------
    buildtoolver: str
        The version of the build tool used to create a package
    """

    buildtoolver: str


class Installed(BaseModel):
    """A list of package names and versions installed during the creation of a package.

    Attributes
    ----------
    installed: list[str]
        A list of strings representing <package_name>-<epoch><version>-<pkgrel>-<architecture> of packages installed
        during the creation of a package
    """

    installed: list[  # type: ignore[valid-type]
        constr(pattern=rf"^({PACKAGE_NAME})-({EPOCH}|){VERSION}-{PKGREL}-{ARCHITECTURE}$")  # noqa: F722
    ]

    @classmethod
    def as_models(cls, installed: list[str]) -> list[tuple[PkgName, PkgVer, ArchitectureEnum]]:
        """Return a list of installed dependencies as models.

        Parameters
        ----------
        installed: list[str]
            A list of strings, each describing pkgname-version-architecture

        Raises
        ------
        ValidationError
            If any of the entries in installed can not be validated using PkgName, PkgVer, Architecture

        Returns
        -------
        list[tuple[PkgName, PkgVer, ArchitectureEnum]]
            A list of PkgName, PkgVer and ArchitectureEnum tuples, which describe each entry in installed
        """
        return [
            (
                PkgName(pkgname="-".join(dep.split("-")[0:-3])),
                PkgVer(pkgver="-".join(dep.split("-")[-3:-1])),
                ArchitectureEnum(dep.split("-")[-1]),
            )
            for dep in installed
        ]


class PkgArch(BaseModel):
    """A CPU architecture for a package.

    Refer to the arch subsection in https://man.archlinux.org/man/PKGBUILD.5.en#OPTIONS_AND_DIRECTIVES for details

    Attributes
    ----------
    pkgarch: str
        A valid CPU architecture for a package
    """

    pkgarch: constr(pattern=rf"^{ARCHITECTURE}$")  # type: ignore[valid-type]  # noqa: F722


class PkgBuildSha256Sum(BaseModel):
    """A SHA-256 checksum for a PKGBUILD file of a package.

    Attributes
    ----------
    pkgbuild_sha256sum: str
        A string representing a SHA-256 checksum for a PKGBUILD of a package
    """

    pkgbuild_sha256sum: constr(pattern=rf"{SHA256}")  # type: ignore[valid-type]


class PkgVer(BaseModel):
    """A version string for a package, consisting of epoch, pkgver and pkgrel.

    Refer to the epoch, pkgrel and pkgver sections in https://man.archlinux.org/man/PKGBUILD.5.en#OPTIONS_AND_DIRECTIVES
    for details

    Attributes
    ----------
    pkgver: str
        A valid package version string which includes epoch, version and pkgrel
    """

    pkgver: constr(pattern=rf"^({EPOCH}|){VERSION}-{PKGREL}$")  # type: ignore[valid-type]  # noqa: F722


class StartDir(BaseModel):
    """An absolute directory used as startdir for a package.

    Refer to the startdir subsection in https://man.archlinux.org/man/PKGBUILD.5.en#PACKAGING_FUNCTIONS

    Attributes
    ----------
    startdir: str
        A string representing the absolute startdir directory of a package
    """

    startdir: str

    @field_validator("startdir")
    @classmethod
    def validate_startdir(cls, startdir: str) -> str:
        """Validate the startdir attribute.

        The startdir attribute may not contain strings that represent absolute Paths or Paths in the home directory

        Parameters
        ----------
        startdir: str
            A string, representing a path

        Returns
        -------
        str
            A validated string, representing an absolute Path
        """
        path = Path(startdir)
        if not path.is_absolute():
            raise ValueError(f"A relative path as startdir is not valid: {path}")

        return startdir


class BuildInfo(BaseModel):
    """The representation of a .BUILDINFO file.

    This is a template class and should not be used directly. Instead instantiate one of the classes derived from it.
    """

    @classmethod
    def from_file(cls, data: StringIO) -> BuildInfo:
        """Create an instance of BuildInfo from an io.StringIO representing the contents of a BUILDINFO file.

        Parameters
        ----------
        data: io.StringIO
            A text stream representing the contents of a .BUILDINFO file

        Returns
        -------
        BuildInfo
            An instance of BuildInfo
        """
        entries: dict[str, int | str | list[str]] = {}

        for line in data:
            [key, value] = [x.strip() for x in line.strip().split("=")]

            assignment_key = BUILDINFO_ASSIGNMENTS.get(key)
            if isinstance(assignment_key, tuple):
                match assignment_key[1]:
                    case FieldTypeEnum.INT:
                        entries[assignment_key[0]] = int(value)
                    case FieldTypeEnum.STRING:
                        entries[assignment_key[0]] = str(value)
                    case FieldTypeEnum.STRING_LIST:
                        entry = entries.get(assignment_key[0])
                        if entry is not None and isinstance(entry, list):
                            entry.append(str(value))
                        else:
                            entries[assignment_key[0]] = [str(value)]
                    # NOTE: the catch all can never be reached but is here to satisfy our tooling
                    case _:  # pragma: no cover
                        continue

        match entries.get("schema_version"):
            case 1:
                return BuildInfoV1.model_validate(entries)
            case 2:
                return BuildInfoV2.model_validate(entries)
            case _:
                raise RepoManagementError(
                    f"Encountered unhandled .BUILDINFO schema_version {entries.get('schema_version')}!"
                )


class BuildInfoV1(
    BuildDate,
    BuildDir,
    BuildEnv,
    BuildInfo,
    Installed,
    Options,
    Packager,
    PkgArch,
    PkgBase,
    PkgBuildSha256Sum,
    PkgName,
    PkgVer,
    SchemaVersionV1,
):
    """The representation of a .BUILDINFO file (version 1).

    Attributes
    ----------
    builddate: NonNegativeInt
        A number >= 0
    builddir: str
        A string representing an absolute directory
    buildenv: list[str]
        A list of strings as described by makepkg.conf's BUILDENV option
    installed: list[str]
        A list of strings representing <package_name>-<epoch><version>-<pkgrel>-<architecture> of packages installed
        during the creation of a package
    options: list[str]
        A list of strings representing makepkg.conf OPTIONS used during the creation of a package
    packager: str
        A string representing a packager UID (e.g. "First Last <mail@example.tld>")
    pkgarch: str
        A valid CPU architecture for a package
    pkgbase: str
        A string representing a valid pkgbase for a package
    pkgbuild_sha256sum: str
        A string representing a SHA-256 checksum for a PKGBUILD of a package
    pkgname: str
        A string representing a valid pkgname of a package
    pkgver: str
        A valid package version string which includes epoch, version and pkgrel
    schema_version: int
        1 - representing `format = 1` in the .BUILDINFO file; schema_version is chosen for uniformity
    """

    pass


class BuildInfoV2(
    BuildDate,
    BuildDir,
    BuildEnv,
    BuildInfo,
    BuildTool,
    BuildToolVer,
    Installed,
    Options,
    Packager,
    PkgArch,
    PkgBase,
    PkgBuildSha256Sum,
    PkgName,
    PkgVer,
    SchemaVersionV2,
    StartDir,
):
    """The representation of a .BUILDINFO file (version 2).

    Attributes
    ----------
    builddate: NonNegativeInt
        A number >= 0
    builddir: str
        A string representing an absolute directory
    buildenv: list[str]
        A list of strings as described by makepkg.conf's BUILDENV option
    buildtool: str
        The package name of the build tool used to create a package
    buildtoolver: str
        The version of the build tool used to create a package
    installed: list[str]
        A list of strings representing <package_name>-<epoch><version>-<pkgrel>-<architecture> of packages installed
        during the creation of a package
    options: list[str]
        A list of strings representing makepkg.conf OPTIONS used during the creation of a package
    packager: str
        A string representing a packager UID (e.g. "First Last <mail@example.tld>")
    pkgarch: str
        A valid CPU architecture for a package
    pkgbase: str
        A string representing a valid pkgbase for a package
    pkgbuild_sha256sum: str
        A string representing a SHA-256 checksum for a PKGBUILD of a package
    pkgname: str
        A string representing a valid pkgname of a package
    pkgver: str
        A valid package version string which includes epoch, version and pkgrel
    schema_version: int
        2 - representing `format = 2` in the .BUILDINFO file; schema_version is chosen for uniformity
    startdir: str
        A string representing the absolute startdir directory of a package
    """

    @model_validator(mode="before")
    def validate_devtools_version(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate the use of a valid version string when devtools is the buildtool.

        Parameters
        ----------
        values: dict[str, Any]
            A dict with all values of a BuildInfoV2 object

        Raises
        ------
        ValueError
            If buildtool is "devtools" and buildtoolver does not follow the format
            <optional_epoch><pkgver>-<pkgrel>-<arch>

        Returns
        -------
        values: dict[str, Any]
            The unmodified dict with all values of a BuildInfoV2 instance
        """
        buildtool, buildtoolver = str(values.get("buildtool")), str(values.get("buildtoolver"))
        if buildtool == "devtools" and not fullmatch(rf"^({EPOCH}|){VERSION}-{PKGREL}-{ARCHITECTURE}$", buildtoolver):
            raise ValueError(
                "When building with devtools the buildtoolver must be of the format "
                f"<optional_epoch><pkgver>-<pkgrel>-<arch>, but it is {buildtoolver}!"
            )

        return values


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema of selected pydantic models to an output directory.

    Parameters
    ----------
    output: Path
        A path to which to output the JSON schema files

    Raises
    ------
    RuntimeError
        If output is not an existing directory
    """
    classes = [BuildInfoV1, BuildInfoV2]

    if isinstance(output, str):
        output = Path(output)

    if not output.exists():
        raise RuntimeError(f"The output directory {output} must exist!")

    for class_ in classes:
        with open(output / f"{class_.__name__}.json", "wb") as f:
            f.write(
                dumps(
                    class_.model_json_schema(),  # type: ignore[attr-defined]
                    option=OPT_INDENT_2 | OPT_APPEND_NEWLINE | OPT_SORT_KEYS,
                )
            )



"""File handling in repod."""

from pathlib import Path

from repod.files.buildinfo import BuildInfo  # noqa: F401
from repod.files.buildinfo import export_schemas as buildinfo_export_schemas
from repod.files.common import extract_file_from_tarfile, open_tarfile  # noqa: F401
from repod.files.mtree import MTree, MTreeEntry  # noqa: F401
from repod.files.mtree import export_schemas as mtree_export_schemas
from repod.files.package import Package  # noqa: F401
from repod.files.package import export_schemas as package_export_schemas
from repod.files.pkginfo import PkgInfo  # noqa: F401
from repod.files.pkginfo import export_schemas as pkginfo_export_schemas
from repod.files.srcinfo import PkgBaseSection, PkgNameSection, SrcInfo  # noqa: F401
from repod.files.srcinfo import export_schemas as srcinfo_export_schemas


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema files of the repod.files package to a directory.

    Parameters
    ----------
    output: Path
        A directory to write the JSON schema files to
    """
    buildinfo_export_schemas(output=output)
    mtree_export_schemas(output=output)
    package_export_schemas(output=output)
    pkginfo_export_schemas(output=output)
    srcinfo_export_schemas(output=output)



"""Pydantic models shared throughout the codebase."""

from __future__ import annotations

from pathlib import Path

from email_validator import EmailNotValidError, validate_email
from pydantic import (
    BaseModel,
    HttpUrl,
    NonNegativeInt,
    PositiveInt,
    conint,
    constr,
    field_validator,
)

from repod.common.regex import (
    ARCHITECTURE,
    BASE64,
    EPOCH,
    MD5,
    OPTIONS,
    PACKAGE_FILENAME,
    PACKAGE_NAME,
    PACKAGER_NAME,
    PKGREL,
    SHA256,
    VERSION,
)
from repod.version import alpm


class Arch(BaseModel):
    """A model describing a single 'arch' attribute.

    Attributes
    ----------
    arch: str
        The attribute can be used to describe the (required) data below an %ARCH% identifier in a 'desc' file, which
        identifies a package's architecture
    """

    arch: constr(pattern=f"^{ARCHITECTURE}$")  # type: ignore[valid-type]  # noqa: F722


class Backup(BaseModel):
    """A model describing a single 'backup' attribute.

    Attributes
    ----------
    backup: list[str] | None
        The attribute can be used to describe the (optional) data below a %BACKUP% identifier in a 'desc' file, which
        identifies which file(s) of a package pacman will create backups for
    """

    backup: list[str] | None = None

    @field_validator("backup")
    @classmethod
    def validate_backup(cls, backup: list[str] | None) -> list[str] | None:  # noqa: N805
        """Validate the backup attribute.

        The backup attribute may not contain strings that represent absolute Paths or Paths in the home directory

        Parameters
        ----------
        backup: list[str] | None
            An optional list of strings, representing paths to validate

        Returns
        -------
        list[str] | None
            None if backup is None, empty list if backup is empty list, or a validated list of strings
        """
        if backup:
            for file in backup:
                path = Path(file)
                if path.is_absolute():
                    raise ValueError(f"Absolute paths in a list of files are not valid: {path}")
                parts = path.parts
                if parts[0] == "home" and len(parts) > 1:
                    raise ValueError(
                        f"Files or directories in the home directory is not valid in a list of files: {path}"
                    )

        return backup


class Base(BaseModel):
    """A model describing a single 'base' attribute.

    Attributes
    ----------
    base: str
        The attribute can be used to describe the (required) data below a %BASE% identifier in a 'desc' file, which
        identifies a package's pkgbase
    """

    base: constr(pattern=f"^{PACKAGE_NAME}$")  # type: ignore[valid-type]  # noqa: F722


class BuildDate(BaseModel):
    """A model describing a single 'builddate' attribute.

    Attributes
    ----------
    builddate: int
        The attribute can be used to describe the (required) data below a %BUILDDATE% identifier in a 'desc' file,
        which identifies a package's build date (represented in seconds since the epoch)
    """

    builddate: NonNegativeInt


class CheckDepends(BaseModel):
    """A model describing a single 'checkdepends' attribute.

    Attributes
    ----------
    checkdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %CHECKDEPENDS% identifier in a 'desc' file,
        which identifies a package's checkdepends
    """

    checkdepends: list[str] | None = None


class Conflicts(BaseModel):
    """A model describing a single 'conflicts' attribute.

    Attributes
    ----------
    conflicts: list[str] | None
        The attribute can be used to describe the (optional) data below a %CONFLICTS% identifier in a 'desc' file, which
        identifies what other package(s) a package conflicts with
    """

    conflicts: list[str] | None = None


class CSize(BaseModel):
    """A model describing a single 'csize' attribute.

    Attributes
    ----------
    csize: int
        The attribute can be used to describe the (required) data below a %CSIZE% identifier in a 'desc' file, which
        identifies a package's size
    """

    csize: NonNegativeInt


class Depends(BaseModel):
    """A model describing a single 'depends' attribute.

    Attributes
    ----------
    depends: list[str] | None
        The attribute can be used to describe the (optional) data below a %DEPENDS% identifier in a 'desc' file, which
        identifies what other package(s) a package depends on
    """

    depends: list[str] | None = None


class Desc(BaseModel):
    """A model describing a single 'desc' attribute.

    Attributes
    ----------
    desc: str
        The attribute can be used to describe the (required) data below a %DESC% identifier in a 'desc' file, which
        identifies a package's description
    """

    desc: str


class FileName(BaseModel):
    """A model describing a single 'filename' attribute.

    Attributes
    ----------
    filename: str
        The attribute can be used to describe the (required) data below a %FILENAME% identifier in a 'desc' file, which
        identifies a package's file name
    """

    filename: constr(pattern=f"^{PACKAGE_FILENAME}$")  # type: ignore[valid-type]  # noqa: F722


class FileList(BaseModel):
    """A model describing an optional list of files.

    Attributes
    ----------
    files: list[str] | None
        The attribute can be used to describe the (optional) data below a %FILES% identifier in a 'files' file, which
        identifies which file(s) belong to a package
    """

    files: list[str] | None = None

    @field_validator("files")
    @classmethod
    def validate_files(cls, files: list[str] | None) -> list[str] | None:  # noqa: N805
        """Validate the files attribute.

        The files attribute may not contain strings that represent absolute Paths or Paths in the home directory

        Parameters
        ----------
        files: list[str] | None
            An optional list of strings, representing paths to validate

        Returns
        -------
        list[str] | None
            None if files is None, empty list if files is empty list, or a validated list of strings
        """
        if files:
            for file in files:
                path = Path(file)
                if path.is_absolute():
                    raise ValueError(f"Absolute paths in a list of files are not valid: {path}")
                parts = path.parts
                if parts[0] == "home" and len(parts) > 1:
                    raise ValueError(
                        f"Files or directories in the home directory is not valid in a list of files: {path}"
                    )

        return files


class Groups(BaseModel):
    """A model describing a single 'groups' attribute.

    Attributes
    ----------
    groups: list[str] | None
        The attribute can be used to describe the (optional) data below a %GROUPS% identifier in a 'desc' file, which
        identifies a package's groups
    """

    groups: list[constr(pattern=f"^{PACKAGE_NAME}$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class ISize(BaseModel):
    """A model describing a single 'isize' attribute.

    Attributes
    ----------
    isize: int
        The attribute can be used to describe the (required) data below an %ISIZE% identifier in a 'desc' file, which
        identifies a package's installed size
    """

    isize: NonNegativeInt


class License(BaseModel):
    """A model describing a single 'license' attribute.

    Attributes
    ----------
    license: list[str]
        The attribute can be used to describe the (required) data below a %LICENSE% identifier in a 'desc' file, which
        identifies a package's license(s)
    """

    license: list[str]  # noqa: A003


class MakeDepends(BaseModel):
    """A model describing a single 'makedepends' attribute.

    Attributes
    ----------
    makedepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %MAKEDEPENDS% identifier in a 'desc' file,
        which identifies a package's makedepends
    """

    makedepends: list[str] | None = None


class Md5Sum(BaseModel):
    """A model describing a single 'md5sum' attribute.

    Attributes
    ----------
    md5sum: str
        The attribute can be used to describe the (required) data below an %MD5SUM% identifier in a 'desc' file, which
        identifies a package's md5 checksum
    """

    md5sum: constr(pattern=MD5)  # type: ignore[valid-type]


class Name(BaseModel):
    """A model describing a single 'name' attribute.

    Attributes
    ----------
    name: str
        The attribute can be used to describe the (required) data below a %NAME% identifier in a 'desc' file, which
        identifies a package's name
    """

    name: constr(pattern=f"^{PACKAGE_NAME}$")  # type: ignore[valid-type]  # noqa: F722


class Options(BaseModel):
    """A list of makepkg.conf OPTIONS used during the creation of a package.

    For valid values refer to the OPTIONS subsection in https://man.archlinux.org/man/makepkg.conf.5#OPTIONS

    Attributes
    ----------
    options: list[str] | None
        An optional list of strings representing makepkg.conf OPTIONS used during the creation of a package
    """

    options: list[constr(pattern=rf"^{OPTIONS}$")] | None = None  # type: ignore[valid-type]  # noqa: F722


class Packager(BaseModel):
    """A model describing a single 'packager' attribute.

    Attributes
    ----------
    packager: str
        The attribute can be used to describe the (required) data below a %PACKAGER% identifier in a 'desc' file, which
        identifies a package's packager
    """

    packager: constr(pattern=(rf"^{PACKAGER_NAME}\s<(.*)>$"))  # type: ignore[valid-type]  # noqa: F722

    @field_validator("packager")
    @classmethod
    def validate_packager_has_valid_email(cls, packager: str) -> str:  # noqa: N805
        """Validate that Packager has an email in the UID.

        Parameters
        ----------
        packager: str
            The packager UID string

        Raises
        ------
        ValuError
            If no valid mail is found in packager

        Returns
        -------
        str
            A validated Packager UID string
        """
        email = packager.replace(">", "").split("<")[1]
        try:
            validate_email(email, check_deliverability=False)
        except EmailNotValidError as e:
            raise ValueError(f"The packager email is not valid: {email}\n{e}") from e

        return packager


class PgpSig(BaseModel):
    """A model describing a single 'pgpsig' attribute.

    Attributes
    ----------
    pgpsig: str
        The attribute can be used to describe the (optional) data below a %PGPSIG% identifier in a 'desc' file, which
        identifies a package's PGP signature
    """

    pgpsig: constr(pattern=f"^{BASE64}$") | None = None  # type: ignore[valid-type]  # noqa: F722


class PkgBase(BaseModel):
    """A pkgbase for a package.

    Refer to https://man.archlinux.org/man/PKGBUILD.5.en#PACKAGE_SPLITTING for details on pkgbase

    Attributes
    ----------
    pkgbase: str
        A string representing the pkgbase of a package
    """

    pkgbase: constr(pattern=rf"^{PACKAGE_NAME}$")  # type: ignore[valid-type]  # noqa: F722


class PkgDesc(BaseModel):
    """A model describing a single pkgdesc attribute.

    Attributes
    ----------
    pkgdesc: str
        A string used as package description
    """

    pkgdesc: str


class PkgName(BaseModel):
    """A pkgname of a package.

    Refer to the pkgname section in https://man.archlinux.org/man/PKGBUILD.5.en#OPTIONS_AND_DIRECTIVES for details

    Attributes
    ----------
    pkgname: str
        A string representing the pkgname of a package
    """

    pkgname: constr(pattern=rf"^{PACKAGE_NAME}$")  # type: ignore[valid-type]  # noqa: F722


class Provides(BaseModel):
    """A model describing a single 'provides' attribute.

    Attributes
    ----------
    provides: list[str] | None
        The attribute can be used to describe the (optional) data below a %PROVIDES% identifier in a 'desc' file, which
        identifies what other package(s) a package provides
    """

    provides: list[str] | None = None


class Replaces(BaseModel):
    """A model describing a single 'replaces' attribute.

    Attributes
    ----------
    replaces: list[str] | None
        The attribute can be used to describe the (optional) data below a %REPLACES% identifier in a 'desc' file, which
        identifies what other package(s) a package replaces
    """

    replaces: list[str] | None = None


class SchemaVersionV1(BaseModel):
    """A model describing a schema version 1.

    Attributes
    ----------
    schema_version: PositiveInt
        A schema version - 1 - for a model
    """

    schema_version: conint(ge=1, le=1) = 1  # type: ignore[valid-type]


class SchemaVersionV2(BaseModel):
    """A model describing a schema version 2.

    Attributes
    ----------
    schema_version: PositiveInt
        A schema version - 2 - for a model
    """

    schema_version: conint(ge=2, le=2) = 2  # type: ignore[valid-type]


class Sha256Sum(BaseModel):
    """A model describing a single 'sha256sum' attribute.

    Attributes
    ----------
    sha256sum: str
        The attribute can be used to describe the (required) data below an %SHA256SUM% identifier in a 'desc' file,
        which identifies a package's sha256 checksum
    """

    sha256sum: constr(pattern=SHA256)  # type: ignore[valid-type]


class OptDepends(BaseModel):
    """A model describing a single 'optdepends' attribute.

    Attributes
    ----------
    optdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %OPTDEPENDS% identifier in a 'desc' file,
        which identifies what other package(s) a package optionally depends on
    """

    optdepends: list[str] | None = None


class Url(BaseModel):
    """A model describing a single 'url' attribute.

    Attributes
    ----------
    url: str
        The attribute can be used to describe the (required) data below a %URL% identifier in a 'desc' file, which
        identifies a package's URL
    """

    url: HttpUrl


class SourceUrl(BaseModel):
    """A URL pointing at sources.

    Attributes
    ----------
    source_url: HttpUrl | None
        An optional url that points at sources (defaults to None)
    """

    source_url: HttpUrl | None = None


class Epoch(BaseModel):
    """A model dscribing a single 'epoch' attribute.

    The epoch denotes a downgrade in version of a given package (a version with an epoch trumps one without)

    Attributes
    ----------
    epoch: PositiveInt | None
        An optional positive integer representing the epoch of a package
    """

    epoch: PositiveInt | None = None


class PkgRel(BaseModel):
    """A model dscribing a single 'pkgrel' attribute.

    The pkgrel denotes the build version of a given package

    Attributes
    ----------
    pkgrel: str
        A string representing the pkgrel (package release version) of a package
    """

    pkgrel: constr(pattern=rf"^{PKGREL}$")  # type: ignore[valid-type]  # noqa: F722


class PkgVer(BaseModel):
    """A model dscribing a single 'pkgver' attribute.

    The pkgver denotes the upstream version of a given package

    Attributes
    ----------
    pkgver: str
        A string representing the pkgver (upstream package version) of a package
    """

    pkgver: constr(pattern=rf"^({VERSION})$")  # type: ignore[valid-type]  # noqa: F722


class Version(BaseModel):
    """A model describing a single 'version' attribute.

    Attributes
    ----------
    version: str
        The attribute can be used to describe the (required) data below a %VERSION% identifier in a 'desc' file, which
        identifies a package's version (this is the accumulation of epoch, pkgver and pkgrel)
    """

    version: constr(pattern=rf"^({EPOCH}|){VERSION}-{PKGREL}$")  # type: ignore[valid-type]  # noqa: F722

    def get_epoch(self: Version) -> Epoch | None:
        """Return the epoch of the version.

        Returns
        -------
        int | None
            An optional string representing the epoch of the version
        """
        if ":" in self.version:
            return Epoch(epoch=self.version.split(":")[0])

        return None

    def get_pkgver(self: Version) -> PkgVer:
        """Return the pkgver of the version.

        Returns
        -------
        PkgVer
            A PkgVer representing the pkgver of the version
        """
        pkgver_pkgrel = self.version.split(":")[1] if ":" in self.version else self.version
        return PkgVer(pkgver=str(pkgver_pkgrel.split("-")[0]))

    def get_pkgrel(self: Version) -> PkgRel:
        """Return the pkgrel of the version.

        Returns
        -------
        PkgRel
            A PkgRel representing the pkgrel of the version
        """
        pkgver_pkgrel = self.version.split(":")[1] if ":" in self.version else self.version
        return PkgRel(pkgrel=str(pkgver_pkgrel.split("-")[1]))

    def vercmp(self: Version, version: Version) -> int:
        """Compare the version with another.

        The comparison algorithm is based on pyalpm's/ pacman's vercmp behavior.
        If PYALPM_VERCMP is True, pyalpm has been imported and its implementation of vercmp() is used.

        Returns
        -------
        int
            -1 if self.version is older than version
            0 if self.version is equal to version
            1 if self.version is newer than version
        """
        return alpm.pkg_vercmp(self.version, version.version)

    def is_older_than(self: Version, version: Version) -> bool:
        """Check whether the version is older than a provided version.

        Parameters
        ----------
        version: Version
            Another version to compare that of self to

        Returns
        -------
        True if self.version is older than the provided version, False otherwise.
        """
        return True if self.vercmp(version=version) < 0 else False

    def is_newer_than(self: Version, version: Version) -> bool:
        """Check whether the version is newer than a provided version.

        Parameters
        ----------
        version: Version
            Another version to compare that of self to

        Returns
        -------
        True if self.version is newer than the provided version, False otherwise.
        """
        return True if self.vercmp(version=version) > 0 else False



"""Regular expressions used for matching strings in packages and sync databases."""

from repod.common.enums import (
    ArchitectureEnum,
    tar_compression_types_for_filename_regex,
)

RELATIVE_MTREE_PATH = r"[A-Za-z0-9.,:;/_()@\\&$?!+%~{}<>*\-\"\'\[\]\`^|]+"
"""
RELATIVE_MTREE_PATH may include printable ASCII characters other than space, #, =. The backslash is included as it's a
result of encoding special characters. See the mtree implementation in libarchive:
https://github.com/libarchive/libarchive/blob/v3.6.1/libarchive/archive_write_set_format_mtree.c#L307
"""
ABSOLUTE_MTREE_PATH = rf"/{RELATIVE_MTREE_PATH}"
ARCHITECTURE = rf"({ArchitectureEnum.as_or_regex()})"
B2 = r"^[a-f0-9]{128}$"
BASE64 = r"[0-9A-Za-z/+]+={0,2}"
BUILDENVS = r"(!|)[\w\-.]+"
CK = r"^[0-9]{10}$"
EPOCH = r"[1-9]+[0-9]*:"
MD5 = r"^[a-f0-9]{32}$"
OPTIONS = r"(!|)[\w\-.]+"
PGP_KEY_ID = r"^[A-F0-9]{40}$"
PKGREL = r"[1-9]+[0-9]*(|[.]{1}[1-9]+[0-9]*)"
PACKAGE_NAME = r"[a-z\d_@+]+[a-z\d\-._@+]*"
PACKAGER_NAME = r"[\w\s\-().]+"
SHA1 = r"^[a-f0-9]{40}$"
SHA224 = r"^[a-f0-9]{56}$"
SHA256 = r"^[a-f0-9]{64}$"
SHA384 = r"^[a-f0-9]{96}$"
SHA512 = r"^[a-f0-9]{128}$"
VERSION = r"([A-Za-z\d]+)[_+.]?[A-Za-z\d_+.]*"
PACKAGE_FILENAME = (
    rf"{PACKAGE_NAME}-({EPOCH}|){VERSION}-{PKGREL}-{ARCHITECTURE}"
    rf"(.pkg.tar)({tar_compression_types_for_filename_regex()})"
)
SIGNATURE_FILENAME = rf"{PACKAGE_FILENAME}(.sig)"



"""Custom Exceptions for repod."""


class RepoManagementError(Exception):
    """A class of Errors that is raised on issues with handling a repository database."""


class RepoManagementFileError(RepoManagementError):
    """An Error that is raised on issues with reading or writing files using repo_managment."""


class RepoManagementValidationError(RepoManagementError):
    """An Error that is raised on issues with validating files using repo_managment."""


class RepoManagementFileNotFoundError(RepoManagementFileError, FileNotFoundError):
    """An Error that is raised when a file can not be found."""


class TaskError(RepoManagementError):
    """An Error that is raised when an error occurs in a Task."""


class FileParserError(RepoManagementError):
    """An Error that is raised when an error occurs during parsing of a file."""



"""Functions for handling repod's CLI."""

import asyncio
from argparse import ArgumentParser, Namespace
from logging import DEBUG, INFO, WARNING, StreamHandler, debug, getLogger
from pathlib import Path
from sys import exit, stderr, stdout
from unittest.mock import patch

from orjson import dumps

from repod import export_schemas
from repod.action.workflow import (
    add_packages,
    add_packages_dryrun,
    write_sync_databases,
)
from repod.cli import argparse
from repod.common.enums import RepoDirTypeEnum, RepoTypeEnum
from repod.config import SystemSettings, UserSettings
from repod.config.defaults import ORJSON_OPTION
from repod.files import Package
from repod.repo import SyncDatabase


def exit_on_error(message: str, argparser: ArgumentParser | None = None) -> None:
    """Print a message to stderr, optionally print argparse help and exit with return code 1.

    Parameters
    ----------
    message: str
        A message to print to stderr
    argparser: ArgumentParser | None
        An optional Argumentparser on which to call print_help()
    """
    print(message, file=stderr)
    if argparser:
        argparser.print_help()
    exit(1)


def repod_file_package(args: Namespace, settings: SystemSettings | UserSettings) -> None:
    """Package related actions from the repod-file script.

    Parameters
    ----------
    args: Namespace
        The options used for the Package related actions
    settings: SystemSettings | UserSettings
        A Settings instance that is used for deriving repository directories from

    Raises
    ------
    RuntimeError
        If a debug repository is targetted, but any of the supplied packages is not a debug package (can only be
        successfully checked if PkgInfoV2 is used in the package).
        If an invalid subcommand is provided.
    """
    pretty = ORJSON_OPTION if hasattr(args, "pretty") and args.pretty else 0
    match args.package:
        case "inspect":
            for package_path in args.file:
                model = asyncio.run(
                    Package.from_file(
                        package=package_path,
                        signature=Path(str(package_path) + ".sig") if args.with_signature else None,
                    )
                )

                if args.buildinfo:
                    print(
                        dumps(
                            model.buildinfo.model_dump(mode="json"),  # type: ignore[attr-defined]
                            option=pretty,
                        ).decode("utf-8")
                    )
                elif args.mtree:
                    print(
                        dumps(
                            model.mtree.model_dump(mode="json"),  # type: ignore[attr-defined]
                            option=pretty,
                        ).decode("utf-8")
                    )
                elif args.pkginfo:
                    print(
                        dumps(
                            model.pkginfo.model_dump(mode="json"),  # type: ignore[attr-defined]
                            option=pretty,
                        ).decode("utf-8")
                    )
                else:
                    print(dumps(model.model_dump(mode="json"), option=pretty).decode("utf-8"))
        case _:
            exit_on_error(
                message="No subcommand provided to the 'package' command!\n",
                argparser=argparse.ArgParseFactory.repod_file(),
            )


def repod_file_repo_importpkg(args: Namespace, settings: SystemSettings | UserSettings) -> None:
    """Import a package (optionally with signature file) to a repository and write its sync databases.

    Parameters
    ----------
    args: Namespace
        The options used for repo related actions
    settings: SystemSettings | UserSettings
        A Settings instance that is used for deriving repository directories from
    """
    if args.dry_run:
        add_packages_dryrun(
            settings=settings,
            files=args.file,
            repo_name=args.name,
            repo_architecture=args.architecture,
            debug_repo=args.debug,
            with_signature=args.with_signature,
            pkgbase_urls=dict(args.source_url),
        )
        return

    add_packages(
        settings=settings,
        files=args.file,
        repo_name=args.name,
        repo_architecture=args.architecture,
        debug_repo=args.debug,
        staging_repo=args.staging,
        testing_repo=args.testing,
        with_signature=args.with_signature,
        pkgbase_urls=dict(args.source_url),
    )


def repod_file_repo(args: Namespace, settings: SystemSettings | UserSettings) -> None:
    """Handle repository related actions from the repod-file script.

    Parameters
    ----------
    args: Namespace
        The options used for repo related actions
    settings: SystemSettings | UserSettings
        A Settings instance that is used for deriving repository directories from

    Raises
    ------
    RuntimeError
        If an invalid subcommand is provided.
    """
    match args.repo:
        case "importdb":
            management_repo_dir = settings.get_repo_path(
                repo_dir_type=RepoDirTypeEnum.MANAGEMENT,
                name=args.name,
                architecture=args.architecture,
                repo_type=RepoTypeEnum.from_bool(
                    debug=args.debug,
                    staging=args.staging,
                    testing=args.testing,
                ),
            )
            for base, outputpackagebase in asyncio.run(
                SyncDatabase(
                    database=args.file,
                    desc_version=settings.syncdb_settings.desc_version,
                    files_version=settings.syncdb_settings.files_version,
                ).outputpackagebases()
            ):
                with open(management_repo_dir / f"{base}.json", "wb") as output_file:
                    output_file.write(dumps(outputpackagebase.model_dump(mode="json"), option=ORJSON_OPTION))
        case "importpkg":
            repod_file_repo_importpkg(args=args, settings=settings)
        case "writedb":
            write_sync_databases(
                settings=settings,
                repo_name=args.name,
                repo_architecture=args.architecture,
                debug_repo=args.debug,
                staging_repo=args.staging,
                testing_repo=args.testing,
            )
        case _:
            exit_on_error(
                message="No subcommand provided to the 'repo' command!\n",
                argparser=argparse.ArgParseFactory.repod_file(),
            )


def repod_file_schema(args: Namespace) -> None:
    """Handle JSON schema related actions from the repod-file script.

    Parameters
    ----------
    args: Namespace
        The options used for the JSON schema related actions

    Raises
    ------
    RuntimeError
        If an invalid subcommand is provided.
    """
    match args.schema:
        case "export":
            export_schemas(output=args.dir)
        case _:
            exit_on_error(
                message="No subcommand provided to the 'schema' command!\n",
                argparser=argparse.ArgParseFactory.repod_file(),
            )


def repod_file() -> None:
    """Delegate calls to repod-file to underlying functions."""
    try:
        args = argparse.ArgParseFactory.repod_file().parse_args()
    except argparse.ArgumentTypeError as e:
        raise RuntimeError(e)

    loglevel = WARNING

    if args.verbose_mode:
        loglevel = INFO
    if args.debug_mode:
        loglevel = DEBUG

    logger = getLogger()
    logger.setLevel(loglevel)
    ch = StreamHandler(stream=stdout)
    ch.setLevel(loglevel)
    logger.addHandler(ch)
    debug(f"ArgumentParser: {args}")

    with patch("repod.config.settings.CUSTOM_CONFIG", args.config):
        settings = SystemSettings() if args.system else UserSettings()
    debug(f"Settings: {settings}")

    match args.subcommand:
        case "package":
            repod_file_package(args=args, settings=settings)  # type: ignore[arg-type]
        case "repo":
            repod_file_repo(args=args, settings=settings)  # type: ignore[arg-type]
        case "schema":
            repod_file_schema(args=args)
        case _:
            exit_on_error(
                message="No subcommand specified!\n",
                argparser=argparse.ArgParseFactory.repod_file(),
            )
